from __future__ import annotations

from typing import List

import firefly_integration.domain as domain


class Database:
    name: str = None
    path: str = None
    description: str = None
    tables: List[domain.Table] = []

    def __init__(self, name: str, path: str, tables: List[domain.Table], description: str = None):
        self.name = name
        self.path = path
        self.tables = tables
        self.description = description

        for table in self.tables:
            table.database = self

    def get_table(self, name: str):
        ret = list(filter(lambda x: x.name == name, self.tables))

        return None if len(ret) == 0 else ret[0]



from .execute_workflow import ExecuteWorkflow
from .function_parameters import FunctionParameters
from .function_registry import FunctionRegistry
from .workflow import *
from .workflow_function import WorkflowFunction



from __future__ import annotations

import inspect

import firefly_di as di

import firefly_integration.infrastructure.service.dal as dal
from firefly_integration.domain.service.dal import Dal

dal_class = None
for k, v in dal.__dict__.items():
    if inspect.isclass(v) and issubclass(v, Dal):
        dal_class = v
        break


class Container(di.Container):
    dal: Dal = dal_class



from __future__ import annotations

from abc import ABC, abstractmethod

import firefly as ff

from .function_parameters import FunctionParameters


class WorkflowFunction(ff.LoggerAware, ff.SystemBusAware, ABC):
    @abstractmethod
    def __call__(self, p: FunctionParameters) -> FunctionParameters:
        pass



from __future__ import annotations

import importlib

import firefly as ff

import firefly_integration.domain as ffi


@ff.on('firefly.ApplicationLayerLoaded')
class LoadDataCatalogs(ff.ApplicationService):
    _context_map: ff.ContextMap = None
    _catalog_registry: ffi.CatalogRegistry = None

    def __call__(self, **kwargs):
        for context in self._context_map.contexts:
            if context.name != 'firefly' and not context.name.startswith('firefly_'):
                self._load_data_catalogs(context)

    def _load_data_catalogs(self, context: ff.Context):
        module_name = f'{context.name}.domain'
        try:
            module = importlib.import_module(module_name.format(context.name))
        except (ModuleNotFoundError, KeyError):
            return []

        for k, v in module.__dict__.items():
            if isinstance(v, ffi.Catalog):
                self._catalog_registry.add_catalog(v)



from __future__ import annotations

from typing import List, Optional

import firefly as ff
import firefly_integration.domain as domain


class CatalogRegistry(ff.DomainService):
    _catalogs: List[domain.Catalog] = []

    def add_catalog(self, catalog: domain.Catalog):
        self._catalogs.append(catalog)

    def get_all_tables(self) -> List[domain.Table]:
        ret = []
        for catalog in self._catalogs:
            for database in catalog.databases:
                for table in database.tables:
                    ret.append(table)

        return ret

    def get_table(self, table_name: str) -> Optional[domain.Table]:
        for catalog in self._catalogs:
            for database in catalog.databases:
                for table in database.tables:
                    if table.name == table_name:
                        return table



from __future__ import annotations

from datetime import datetime
from hashlib import md5
from typing import List

import awswrangler as wr
import firefly as ff
import pandas as pd
from botocore.exceptions import ClientError

import firefly_integration.domain as domain
from firefly_integration.domain.service.dal import Dal

MAX_FILE_SIZE = 1000000000  # ~1GB
PARTITION_LOCK = 'partition-lock-{}'


class AwsDal(Dal):
    _batch_process: ff.BatchProcess = None
    _remove_duplicates: domain.RemoveDuplicates = None
    _mutex: ff.Mutex = None
    _db_created: dict = {}
    _context: str = None
    _bucket: str = None

    def store(self, df: pd.DataFrame, table: domain.Table):
        self._ensure_db_created(table)

        df = df[list(map(lambda c: c.name, table.columns))]

        if 'created_on' in table.type_dict:
            df['created_on'].fillna(datetime.utcnow(), inplace=True)
        if 'updated_on' in table.type_dict:
            df['updated_on'] = datetime.utcnow()

        params = {
            'df': df,
            'path': table.full_path(),
            'dataset': True,
            'database': table.database.name,
            'table': table.name,
            'partition_cols': table.partition_columns,
            'compression': 'snappy',
            'dtype': table.type_dict,
            'schema_evolution': True,
        }

        if table.time_partitioning is not None:
            params['partition_cols'] = ['dt']
            # params['projection_enabled'] = True
            params['regular_partitions'] = True
            fmt = '%Y'
            if table.time_partitioning == 'month':
                fmt += '-%m'
            if table.time_partitioning == 'day':
                fmt += '-%m-%d'

            df['dt'] = pd.to_datetime(df[table.time_partitioning_column]).dt.strftime(fmt)

            # params['projection_types'] = {'dt': 'enum'}
            # params['projection_values'] = {'dt': self._date_partition_range(df, table.time_partitioning)}

        if not df.empty:
            wr.s3.to_parquet(**params)

    def _date_partition_range(self, df: pd.DataFrame, timeframe: str):
        start = df['dt'].min()
        end = df['dt'].max()
        fmt = '%Y'
        if timeframe == 'month':
            fmt += '-%m'
        if timeframe == 'day':
            fmt += '-%m-%d'

        return ','.join(pd.date_range(start, end, freq='MS').strftime(fmt).to_list())

    def load(self, table: domain.Table, criteria: ff.BinaryOp = None) -> pd.DataFrame:
        pass

    def delete(self, criteria: ff.BinaryOp, table: domain.Table):
        pass

    def get_partitions(self, table: domain.Table, criteria: ff.BinaryOp = None) -> List[str]:
        args = {'database': table.database.name, 'table': table.name}
        if criteria is not None:
            args['expression'] = str(criteria)

        try:
            partitions = wr.catalog.get_parquet_partitions(**args)
        except ClientError:
            return []

        return list(map(lambda p: p.replace('s3://', ''), partitions.keys()))

    def wait_for_tmp_files(self, files: list):
        wr.s3.wait_objects_exist(
            list(map(lambda f: f's3://{self._bucket}/{f}', files)),
            delay=1,
            max_attempts=60,
            use_threads=True
        )

    def read_tmp_files(self, files: list) -> pd.DataFrame:
        return wr.s3.read_parquet(list(map(lambda f: f's3://{self._bucket}/{f}', files)), use_threads=True)

    def write_tmp_file(self, file: str, data: pd.DataFrame):
        wr.s3.to_parquet(data, path=f's3://{self._bucket}/{file}')

    def compact(self, table: domain.Table, path: str):
        path = path.rstrip('/')
        if not path.startswith('s3://'):
            path = f's3://{path}'

        ignore = []
        to_delete = []
        key = None
        n = 0
        for k, size in wr.s3.size_objects(path=f'{path}/', use_threads=True).items():
            if k.endswith('.dat.snappy.parquet'):
                if size >= MAX_FILE_SIZE:
                    ignore.append(f'{k.split("/")[-1]}')
                else:
                    key = k
                n += 1
            else:
                to_delete.append(k)

        if len(to_delete) == 0:
            return  # Nothing new to compact

        with self._mutex(PARTITION_LOCK.format(md5(path.encode('utf-8')).hexdigest())):
            if key is None:
                key = f'{path}/{n + 1}.dat.snappy.parquet'

            df = wr.s3.read_parquet(path=path, path_ignore_suffix=ignore, use_threads=True)
            self._remove_duplicates(df, table)
            try:
                df.reset_index(inplace=True)
            except ValueError:
                pass
            wr.s3.to_parquet(df=df, path=key, compression='snappy', dtype=table.type_dict, use_threads=True)
            wr.s3.delete_objects(to_delete, use_threads=True)

    def _ensure_db_created(self, table: domain.Table):
        if table.database.name not in self._db_created:
            wr.catalog.create_database(name=table.database.name, exist_ok=True,
                                       description=table.database.description or '')
            self._db_created[table.database.name] = True



from __future__ import annotations

import importlib
import inspect
from datetime import datetime, date

import firefly as ff
import inflection
from troposphere import Template
from troposphere.glue import Database, DatabaseInput, Table, TableInput, Column, StorageDescriptor, SerdeInfo

import firefly_integration.domain as ffi


@ff.agent.pre_deploy_hook(for_='aws')
class AddGlueDataCatalog(ff.AgentExtension):
    _account_id: str = None

    def __call__(self, template: Template, context: ff.Context, env: str, **kwargs):
        for catalog in self._load_catalogs(context):
            for database in catalog.databases:
                self._add_database(template, database)

    def _add_database(self, template: Template, database: ffi.Database):
        db = template.add_resource(Database(
            inflection.camelize(database.name, uppercase_first_letter=True),
            CatalogId=self._account_id,
            DatabaseInput=DatabaseInput(
                Description=database.description or '',
                Name=database.name
            )
        ))

        path = database.path.rstrip('/')
        if not path.startswith('s3://'):
            path = f's3://{path.lstrip("/")}'
        for table in database.tables:
            self._add_table(template, table, database.name, db, path)

    def _add_table(self, template: Template, table: ffi.Table, db_name: str, db, prefix: str):
        path = f'{prefix}/{table.path.strip("/")}'
        path = f'{path.rstrip("/")}/{table.name}'

        template.add_resource(Table(
            inflection.camelize(table.name, uppercase_first_letter=True),
            CatalogId=self._account_id,
            DatabaseName=db_name,
            TableInput=TableInput(
                Description=table.description or '',
                Name=table.name,
                Parameters={
                    'projection.enabled': False,
                },
                PartitionKeys=[
                    Column(Comment=part.comment or '', Name=part.name, Type=self._athena_type(part.data_type))
                    for part in table.partitions
                ],
                StorageDescriptor=StorageDescriptor(
                    InputFormat='org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat',
                    OutputFormat='org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat',
                    SerdeInfo=SerdeInfo(
                        SerializationLibrary='org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe'
                    ),
                    Location=path,
                    Columns=[
                        Column(Comment=c.comment or '', Name=c.name, Type=self._athena_type(c.data_type))
                        for c in table.columns
                    ]
                )
            ),
            DependsOn=[db]
        ))

    def _athena_type(self, t: type):
        if t is str:
            return 'string'
        if t is int:
            return 'integer'
        if t is float:
            return 'float'
        if t is bool:
            return 'boolean'
        if t is datetime:
            return 'timestamp'
        if t is date:
            return 'date'

    def _load_catalogs(self, context: ff.Context):
        module_name = context.config.get('domain_module', '{}.domain')
        try:
            module = importlib.import_module(module_name.format(context.name))
        except (ModuleNotFoundError, KeyError):
            return []

        ret = []
        for k, v in module.__dict__.items():
            if isinstance(v, ffi.Catalog):
                ret.append(v)

        return ret



from __future__ import annotations

from typing import List

import firefly_integration.domain as domain


class Catalog:
    databases: List[domain.Database] = []

    def __init__(self, databases: List[domain.Database]):
        self.databases = databases

    def get_database(self, name: str = None):
        if name is None and len(self.databases) == 1:
            return self.databases[0]
        ret = list(filter(lambda x: x.name == name, self.databases))

        return None if len(ret) == 0 else ret[0]



from __future__ import annotations

import importlib
import inspect

import firefly as ff
import inflection

from .function_parameters import FunctionParameters
from .function_registry import FunctionRegistry
from .workflow import Workflow
from .workflow_function import WorkflowFunction
from ..errors import WorkflowFunctionError


class ExecuteWorkflow(ff.DomainService):
    _batch_process: ff.BatchProcess = None
    _function_registry: FunctionRegistry = None
    _loaded: list = []

    def __call__(self, workflow: Workflow, data: FunctionParameters):
        if workflow not in self._loaded:
            self._load_workflow_functions(workflow)
            self._loaded.append(workflow)

        for batch in workflow.batch():
            for function in batch:
                # try:
                data = self._execute_function(function, data)
#                 except Exception as e:
#                     raise WorkflowFunctionError(f"""
# Function: {function}
#
# Input:
#
# {data.debug()}
#
# Original Exception:
#
# {str(e)}
#                     """)
        # for batch in workflow.batch():
        #     self._batch_process(self._execute_function, [(x, data) for x in batch])

        return data

    def _execute_function(self, function: str, data: FunctionParameters):
        return self._function_registry.get(function)(data)

    def _load_workflow_functions(self, workflow: Workflow):
        module = importlib.import_module(workflow.module)
        for k, v in module.__dict__.items():
            if inspect.isclass(v) and issubclass(v, WorkflowFunction):
                self._function_registry.register(inflection.underscore(v.__name__), v)



from __future__ import annotations

from datetime import datetime, date

import pandas as pd

import firefly_integration.domain as domain

ALLOWED_TYPES = (str, int, float, bool, datetime, date)


class NoDefault:
    pass


class Column:
    name: str = None
    data_type: type = None
    description: str = None
    required: bool = False
    default = NoDefault
    meta: dict = {}
    table: domain.Table = None

    def __init__(self, name: str, data_type: type, description: str = None, required: bool = False, default=NoDefault,
                 meta: dict = None):
        if data_type not in ALLOWED_TYPES:
            raise domain.InvalidDataType(
                f'Data type {data_type} is not valid for {name}. Allowed types are: {ALLOWED_TYPES}'
            )

        self.name = name
        self.data_type = data_type
        self.description = description
        self.required = required
        self.default = default
        self.meta = meta or {}

    def set_type(self, df: pd.DataFrame):
        df[self.name].astype(str(self.data_type), inplace=True)
        return df

    @property
    def pandas_type(self):
        if self.data_type is int:
            return 'Int64'
        elif self.data_type is str:
            return 'string'
        else:
            return self.data_type




from __future__ import annotations

import firefly as ff
import pandas as pd

import firefly_integration.domain as domain


@ff.command_handler()
class FilterParquet(ff.ApplicationService):
    _catalog_registry: domain.CatalogRegistry = None
    _sanitize_input_data: domain.SanitizeInputData = None
    _filter_parquet: domain.FilterParquet = None
    _batch_process: ff.BatchProcess = None
    _dal: domain.Dal = None

    def __call__(self, files: list, fields: list, criteria: dict, result_file: str, table_name: str, **kwargs):
        criteria = ff.BinaryOp.from_dict(criteria)
        results = self._batch_process(self._do_filter, [(file, fields, criteria) for file in files])
        df = pd.concat(results)
        table = self._catalog_registry.get_table(table_name)
        df = self._sanitize_input_data(df, table)
        self._dal.write_tmp_file(result_file, df)

    def _do_filter(self, path: str, fields: list, criteria: ff.BinaryOp):
        return self._filter_parquet(path, fields, criteria)



from .compact import *
from .filter_parquet import FilterParquet
from .load_data_catalogs import *
from .map_reduce_functions import *
from .query_data import QueryData
from .store_data import StoreData



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List

import firefly as ff
import pandas as pd

from ..data_catalog.table import Table


class Dal(ABC):
    @abstractmethod
    def store(self, data: pd.DataFrame, table: Table):
        pass

    @abstractmethod
    def load(self, table: Table, criteria: ff.BinaryOp = None) -> pd.DataFrame:
        pass

    @abstractmethod
    def delete(self, criteria: ff.BinaryOp, table: Table):
        pass

    @abstractmethod
    def get_partitions(self, table: Table, criteria: ff.BinaryOp = None) -> List[str]:
        pass

    @abstractmethod
    def wait_for_tmp_files(self, files: list):
        pass

    @abstractmethod
    def read_tmp_files(self, files: list) -> pd.DataFrame:
        pass

    @abstractmethod
    def write_tmp_file(self, file: str, data: pd.DataFrame):
        pass

    @abstractmethod
    def compact(self, table: Table, path: str):
        pass



from __future__ import annotations

from io import StringIO

import firefly as ff
import pandas as pd


@ff.query_handler()
class Map(ff.ApplicationService):
    _file_system: ff.FileSystem = None
    _batch_process: ff.BatchProcess = None
    _context: str = None
    _df: pd.DataFrame = None

    def __call__(self, keys: list, fields: list, criteria: dict, types: dict):
        criteria = ff.BinaryOp.from_dict(criteria)
        results = self._batch_process(self._read, [(key[0], fields, criteria, types) for key in keys])

        return pd.concat(results).to_json()

    def _read(self, key: str, fields: list, criteria: ff.BinaryOp, types: dict):
        data = self._file_system.filter(key, fields, criteria)
        ret = pd.read_json(StringIO(data), dtype=types)
        # ret.set_index(['id'], inplace=True)

        return ret



from __future__ import annotations

import graphviz


class Workflow:
    name: str = None

    def __init__(self, name: str = None, module: str = None):
        self.name = name
        self.module = module
        self._nodes = {}

    def step(self, name: str, label: str, dependencies: list = None, url: str = None) -> Dag:
        self._nodes[name] = {
            'dependencies': dependencies or [],
            'label': label,
            'url': url,
        }
        return self

    def batch(self):
        ret = []
        complete = []
        incomplete = list(self._nodes.keys())

        while True:
            batch = []
            for k, v in self._nodes.items():
                if all(e in complete for e in v['dependencies']):
                    if k not in complete:
                        batch.append(k)

            ret.append(batch)
            for item in batch:
                complete.append(item)
                incomplete.remove(item)

            if len(incomplete) == 0:
                break

        return ret

    def render(self, file: str):
        dot = graphviz.Digraph(comment='Pre-Processing of Workouts')

        for k, v in self._nodes.items():
            attrs = {}
            if v['url'] is not None:
                attrs['URL'] = v['url']

            dot.node(k, v['label'], **attrs)
            for dependency in v['dependencies']:
                dot.edge(dependency, k)

        dot.render(file)



from __future__ import annotations

import firefly as ff
import pandas as pd

import firefly_integration.domain as ffi


@ff.query_handler()
class QueryData(ff.ApplicationService):
    _catalog_registry: ffi.CatalogRegistry = None
    _query_warehouse: ffi.QueryWarehouse = None

    def __call__(self, sql: str, **kwargs):
        ret: pd.DataFrame = self._query_warehouse(sql)

        return ret.to_json(orient='records')



from __future__ import annotations

from typing import Type

import firefly as ff
import firefly_di as di

from .workflow_function import WorkflowFunction


class FunctionRegistry(ff.DomainService):
    _container: di.Container = None
    _functions: dict = {}

    def register(self, key: str, function: Type[WorkflowFunction]):
        self._functions[key] = self._container.build(function)

    def get(self, key: str) -> WorkflowFunction:
        return self._functions[key]



from __future__ import annotations

from datetime import datetime, date
from typing import Union, List

import firefly as ff

import firefly_integration.domain as domain
import pandas as pd
import numpy as np


class StoreData(ff.DomainService):
    _sanitize_input_data: domain.SanitizeInputData = None
    _dal: domain.Dal = None

    def __call__(self, data, table: domain.Table):
        df = self._sanitize_input_data(data, table)
        self._dal.store(df, table)



from __future__ import annotations

import firefly as ff
import pandas as pd

import firefly_integration.domain as domain


class RemoveDuplicates(ff.DomainService):
    def __call__(self, df: pd.DataFrame, table: domain.Table):
        try:
            df.sort_values(table.duplicate_sort, inplace=True)
            df.drop_duplicates(subset=table.duplicate_fields, keep='last', inplace=True)
        except IndexError:
            pass



from __future__ import annotations

from datetime import datetime, date
from typing import Union, List

import firefly as ff

import firefly_integration.domain as domain
import pandas as pd
import numpy as np


class MarshalDataframe(ff.DomainService):
    def __call__(self, df: pd.DataFrame, table: domain.Table) -> pd.DataFrame:
        for column in table.columns:
            if column.data_type in (date, datetime) and column.name in df:
                df[column.name] = pd.to_datetime(df[column.name])

        return df



from __future__ import annotations

from datetime import datetime, date
from typing import Union, List

import firefly as ff

import firefly_integration.domain as domain
import pandas as pd
import numpy as np


class SanitizeInputData(ff.DomainService):
    def __call__(self, data: Union[List[dict], dict, pd.DataFrame], table: domain.Table) -> pd.DataFrame:
        if not isinstance(data, pd.DataFrame):
            df = pd.DataFrame(data if isinstance(data, list) else [data])
        else:
            df = data

        for column in table.columns:
            if column.name not in df:
                if column.default is not domain.NoDefault:
                    df[column.name] = column.default
                elif column.required:
                    raise domain.InvalidInputData()
                elif df.index.name != column.name:
                    df[column.name] = np.nan
            if column.data_type in (date, datetime):
                if df[column.name].dtype == 'object':
                    try:
                        df[column.name] = np.float64(df[column.name])
                    except ValueError:
                        pass
                df[column.name] = pd.to_datetime(df[column.name])
            elif df.index.name != column.name:
                if column.data_type is int or column.data_type == 'int':
                    try:
                        df[column.name] = df[column.name].astype(np.float64).astype(np.int64)
                    except ValueError:
                        df[column.name] = np.nan
                else:
                    df[column.name] = df[column.name].astype(column.pandas_type)

        columns = list(map(lambda cc: cc.name, table.columns))
        for c in df.columns:
            if c not in columns:
                try:
                    del df[c]
                except KeyError:
                    pass

        df.replace({np.nan: None}, inplace=True)

        return df



from __future__ import annotations

import uuid
from datetime import datetime, date
from typing import List, Dict, Callable, Union

import inflection
import pandas as pd

import firefly_integration.domain as domain


class Table:
    name: str = None
    path: str = None
    description: str = None
    columns: List[domain.Column] = []
    partitions: List[domain.Column] = []
    duplicate_fields: List[str] = []
    duplicate_sort: List[str] = []
    database: domain.Database = None
    date_grouping: dict = None  # ???
    time_partitioning: str = None
    time_partitioning_column: str = None
    file_name: Callable = None
    _partition_generators: Dict[str, Callable] = None

    def __init__(self, name: str, columns: List[domain.Column], partitions: List[domain.Column] = None, path: str = '',
                 description: str = None, duplicate_fields: List[str] = None, duplicate_sort: List[str] = None,
                 partition_generators: Dict[str, Callable] = None, date_grouping: dict = None,
                 file_name: Callable = None, time_partitioning: str = None, time_partitioning_column: str = None):
        self.name = inflection.tableize(name)
        self.path = path
        self.columns = columns
        self.description = description
        self.partitions = partitions or []
        self.duplicate_fields = duplicate_fields or ['id']
        self.duplicate_sort = duplicate_sort or []
        self._partition_generators = partition_generators
        self.date_grouping = date_grouping
        self.file_name = file_name
        self.time_partitioning = time_partitioning
        self.time_partitioning_column = time_partitioning_column

        for column in self.columns:
            column.table = self

    def get_column(self, name: str):
        for column in self.columns:
            if column.name == name:
                return column
        raise domain.ColumnNotFound(name)

    def generate_partition_path(self, data: pd.DataFrame):
        parts = []
        for partition in self.partitions:
            if self._partition_generators is not None and partition.name in self._partition_generators:
                parts.append(f'{partition.name}={self._partition_generators[partition.name](data)}')
            elif partition.name in data:
                parts.append(f'{partition.name}={data[partition.name]}')
            else:
                raise domain.InvalidPartitionData()

        return '/'.join(parts)

    @property
    def type_dict(self):
        ret = {}
        for column in self.columns:
            ret[column.name] = self._pandas_type(column.data_type)
        return ret

    @property
    def partition_columns(self):
        return list(map(lambda c: c.name, self.partitions))

    def full_path(self, df: pd.DataFrame = None):
        ret = f'{self.database.path}/{self.path or ""}'.rstrip('/')
        ret = f'{ret}/{self.name}'
        if df is not None:
            ret = f'{ret}/{self.generate_partition_path(df)}'
            if self.file_name:
                ret = f'{ret}/{self.file_name(df)}'
            else:
                ret = f'{ret}/{str(uuid.uuid4())}'
        if df is not None:
            ret = f'{ret}.snappy.parquet'

        return ret

    @property
    def time_partition_format(self):
        ret = None
        if self.time_partitioning is not None:
            ret = '%Y'
            if self.time_partitioning == 'month':
                ret += '-%m'
            if self.time_partitioning == 'day':
                ret += '-%m-%d'

        return ret

    def _pandas_type(self, t: type):
        if t is str:
            return 'string'
        if t is int:
            return 'bigint'
        if t is float:
            return 'double'
        if t is bool:
            return 'boolean'
        if t is datetime:
            return 'timestamp'
        if t is date:
            return 'date'



from __future__ import annotations

from typing import Tuple, List

import firefly as ff
from moz_sql_parser import parse

import firefly_integration.domain as domain

OPS = {
    'eq': '==', 'ne': '!=', 'lt': '<', 'gt': '>', 'lte': '<=', 'gte': '>=', 'is': 'is'
}


class SqlParser(ff.DomainService):
    _parts: dict = None
    _non_partition_keys: list = None

    def parse(self, sql: str):
        self._parts = parse(sql)
        self._non_partition_keys = []

    def get_select_fields(self):
        self._ensure_parse_called()
        return self._parts['select']

    def get_table(self):
        self._ensure_parse_called()
        if isinstance(self._parts['from'], str):
            return self._parts['from']
        for table in self._parts['from']:
            if 'name' in table and table['name'] == 'x':
                return table['value']

    def get_sort_order(self) -> Tuple[List[str], List[bool]]:
        self._ensure_parse_called()
        fields = []
        ascending = []
        if 'orderby' in self._parts:
            order_parts = self._parts['orderby']
            if not isinstance(order_parts, list):
                order_parts = [order_parts]
            for part in order_parts:
                fields.append(part['value'])
                if 'sort' in part:
                    ascending.append(part['sort'] == 'asc')
                else:
                    ascending.append(True)
        else:
            fields.append('id')
            ascending.append(True)

        return fields, ascending

    def get_criteria(self, data=None, criteria: dict = None):
        self._ensure_parse_called()

        if data is None:
            if 'where' not in self._parts:
                return
            data = self._parts['where']

        if isinstance(data, str):
            self._non_partition_keys.append(data)
            return f'a:{data}'
        elif isinstance(data, (int, float, bool)):
            return data
        elif isinstance(data, dict) and 'literal' in data:
            return data['literal']

        if isinstance(data, dict):
            for op, value in data.items():
                if op in ('and', 'or'):
                    if criteria is None:
                        criteria = {
                            'l': self.get_criteria(value[0], criteria),
                            'o': op,
                            'r': self.get_criteria(value[1], criteria),
                        }
                        for i in range(2, len(value)):
                            criteria = {
                                'l': criteria,
                                'o': op,
                                'r': self.get_criteria(value[i], criteria)
                            }
                    else:
                        for i in range(len(value)):
                            criteria = {
                                'l': criteria,
                                'o': op,
                                'r': self.get_criteria(value[i], criteria)
                            }

                elif op in ('eq', 'ne', 'lt', 'gt', 'lte', 'gte', 'is'):
                    return {
                        'l': self.get_criteria(value[0], criteria),
                        'o': OPS[op],
                        'r': self.get_criteria(value[1], criteria)
                    }

        return criteria

    def get_all_criteria_attributes(self, partitions: list, criteria: ff.BinaryOp):
        ret = []
        if isinstance(criteria.lhv, ff.Attr) and str(criteria.lhv.attr) not in partitions:
            ret.append(str(criteria.lhv.attr))
        if isinstance(criteria.rhv, ff.Attr) and str(criteria.rhv.attr) not in partitions:
            ret.append(str(criteria.rhv.attr))
        if isinstance(criteria.lhv, ff.BinaryOp):
            ret.extend(self.get_all_criteria_attributes(partitions, criteria.lhv))
        if isinstance(criteria.rhv, ff.BinaryOp):
            ret.extend(self.get_all_criteria_attributes(partitions, criteria.rhv))

        return ret

    def _ensure_parse_called(self):
        if self._parts is None:
            raise domain.IntegrationError('You must call parse() before get_criteria()')



class IntegrationError(Exception):
    pass


class InvalidDataType(IntegrationError):
    pass


class TableNotFound(IntegrationError):
    pass


class ColumnNotFound(IntegrationError):
    pass


class InvalidInputData(IntegrationError):
    pass


class InvalidPartitionData(IntegrationError):
    pass


class WorkflowFunctionError(IntegrationError):
    pass



from __future__ import annotations

import multiprocessing
import uuid
from typing import Optional

import firefly as ff
import pandas as pd

import firefly_integration.domain as domain
import awswrangler as wr


class QueryWarehouse(ff.DomainService):
    _catalog_registry: domain.CatalogRegistry = None
    _sql_parser: domain.SqlParser = None
    _batch_process: ff.BatchProcess = None
    _filter_parquet: FilterParquet = None
    _remove_duplicates: domain.RemoveDuplicates = None
    _dal: domain.Dal = None
    _file_system: ff.FileSystem = None
    _ff_environment: str = None

    def __init__(self):
        self._cpu_count = multiprocessing.cpu_count()
        self._threshold = self._cpu_count

    def __call__(self, sql: str) -> Optional[pd.DataFrame]:
        # This is temporary code to get warehouse queries working. This uses athena. We either need to move this
        # code, specifically the aws wrangler part, to an infrastructure class or finish the original approach using
        # lambda. Also, the database name is assumed here, and it shouldn't be.
        self._sql_parser.parse(sql)
        table: domain.Table = self._catalog_registry.get_table(self._sql_parser.get_table())
        results = wr.athena.read_sql_query(
            sql=sql,
            database=f'data_warehouse_{self._ff_environment}',
            ctas_approach=False,
            use_threads=True
        )

        self._remove_duplicates(results, table)
        self._sort(results)

        return results

        # partition_criteria, select_criteria = self._process_criteria(table)
        # paths = self._dal.get_partitions(table, partition_criteria)
        # if len(paths) == 0:
        #     return None
        #
        # files = self._batch_process(self._list_files, [(path,) for path in paths])
        # files = [item for sb in files for item in sb]
        #
        # fields = self._sql_parser.get_select_fields()
        # if len(files) <= self._threshold:
        #     results = self._batch_process(self._filter_parquet, [(file, fields, select_criteria) for file in files])
        #     results = pd.concat(results)
        # else:
        #     results = self._fan_out(files, fields, select_criteria, table)
        #
        # self._remove_duplicates(results, table)
        # self._sort(results)
        #
        # return results

    def _fan_out(self, files: list, fields: list, select_criteria: ff.BinaryOp, table: domain.Table):
        output_path = f'tmp/ff-query-results/{str(uuid.uuid4())}'
        output_files = []
        for batch in ff.chunk(files, self._threshold):
            output_files.append(f'{output_path}/{str(uuid.uuid4())}.snappy.parquet')
            self.invoke('integration.FilterParquet', {
                'files': batch,
                'fields': fields,
                'criteria': select_criteria.to_dict() if select_criteria is not None else None,
                'result_file': output_files[-1],
                'table_name': table.name,
            }, async_=True)

        return self._wait_for_results(output_files)

    def _process_criteria(self, table: domain.Table):
        criteria_dict = self._sql_parser.get_criteria()
        partition_criteria = None
        select_criteria = None
        if criteria_dict is not None:
            criteria = ff.BinaryOp.from_dict(criteria_dict)
            partitions = list(map(lambda t: t.name, table.partitions))
            partition_criteria = criteria.prune(partitions)
            select_criteria = criteria.prune(
                self._sql_parser.get_all_criteria_attributes(partitions, criteria)
            )
        return partition_criteria, select_criteria

    def _list_files(self, path: str):
        return list(map(lambda f: f[0], self._file_system.list(path)))

    def _wait_for_results(self, files: list):
        self._dal.wait_for_tmp_files(files)

        return self._dal.read_tmp_files(files)

    def _sort(self, data: pd.DataFrame):
        fields, ascending = self._sql_parser.get_sort_order()
        if len(list(map(lambda x: x not in data, fields))) == 0:
            data.sort_values(by=fields, ascending=ascending, inplace=True)


class FilterParquet(ff.DomainService):
    _file_system: ff.FileSystem = None
    _serializer: ff.Serializer = None

    def __call__(self, path: str, fields: list, criteria: ff.BinaryOp):
        return pd.DataFrame(self._serializer.deserialize(
            self._file_system.filter(path.lstrip('s3://'), fields, criteria)
        ))



from __future__ import annotations

import firefly as ff
import firefly_integration.domain as ffi


@ff.command_handler()
class StoreData(ff.ApplicationService):
    _catalog_registry: ffi.CatalogRegistry = None
    _store_data: ffi.StoreData = None

    def __call__(self, data, table: str, **kwargs):
        return self._store_data(data, self._catalog_registry.get_table(table))



from .dal import Dal
from .marshal_dataframe import MarshalDataframe
from .query_warehouse import *
from .remove_duplicates import RemoveDuplicates
from .sanitize_input_data import SanitizeInputData
from .sql_parser import SqlParser
from .store_data import StoreData



from __future__ import annotations

import firefly as ff
import firefly_integration.domain as domain


@ff.command_handler()
class Compact(ff.ApplicationService):
    _dal: domain.Dal = None
    _catalog_registry: domain.CatalogRegistry = None
    _context: str = None

    def __call__(self, table_name: str = None, path: str = None, **kwargs):
        # Run compaction on all tables
        if table_name is None:
            for table in self._catalog_registry.get_all_tables():
                self.invoke(f'{self._context}.Compact', {
                    'table_name': table.name,
                }, async_=True)

        # Run compaction on all partitions for the given table
        elif path is None:
            table = self._catalog_registry.get_table(table_name)
            if len(table.partitions) > 0 or table.time_partitioning is not None:
                for partition in self._dal.get_partitions(table):
                    self.invoke(f'{self._context}.Compact', {
                        'table_name': table.name,
                        'path': partition,
                    }, async_=True)
            else:
                self.invoke(f'{self._context}.Compact', {
                    'table_name': table.name,
                    'path': table.full_path(),
                }, async_=True)

        # Run compaction on single partition
        else:
            table = self._catalog_registry.get_table(table_name)
            self._dal.compact(table=table, path=path)
